import time


class TimeoutError(RuntimeError):
    pass


class TimeOut(object):

    def __init__(self, timeout, interval=1, raise_on_timeout=False):
        self.remaining = timeout
        self.cutoff = time.time() + timeout
        self.interval = interval
        self.timed_out = False
        self.first = True
        self.raise_on_timeout = raise_on_timeout

    def tick(self):
        """Perform a `tick` for this timeout.

        Returns True if we should keep going or False if not.

        Instead of returning False this can raise an exception
        if raise_on_timeout is set.

        """
        self.remaining = self.cutoff - time.time()
        self.timed_out = self.remaining <= 0

        if self.timed_out:
            if self.raise_on_timeout:
                raise TimeoutError()
            else:
                return False

        if self.first:
            self.first = False
        else:
            time.sleep(self.interval)

        return True



"""Low-level Ceph interface.

We expect Ceph Python bindings to be present in the system site packages.
"""

from __future__ import print_function

from ..sysconfig import sysconfig
import hashlib
import json
import logging
import os
import os.path as p
import rados
import rbd
import shutil
import subprocess
import tempfile
import time

logger = logging.getLogger(__name__)


def cmd(cmdline):
    """Execute cmdline with stdin closed to avoid questions on terminal"""
    print(cmdline)
    with open('/dev/null') as null:
        output = subprocess.check_output(cmdline, shell=True, stdin=null)
    # Keep this here for compatibility with tests
    print(output, end='')
    return output


class Volume(object):
    """Low-level manipulation of RBD volumes.

    A volume has a name (shown in `rbd ls`) and a filesystem label. The
    latter is only used for mkfs/mkswap.
    """

    mapped = False

    def __init__(self, ceph, name, label):
        self.ceph = ceph
        self.ioctx = ceph.ioctx
        self.name = name
        self.label = label
        self.rbd = rbd.RBD()
        self.snapshots = Snapshots(self)

    @property
    def fullname(self):
        return self.ioctx.name + '/' + self.name

    @property
    def image(self):
        return rbd.Image(self.ioctx, self.name)

    @property
    def size(self):
        """Image size in Bytes."""
        return self.image.size()

    def exists(self):
        return self.name in self.rbd.list(self.ioctx)

    def ensure_presence(self, size=1024):
        if self.name in self.rbd.list(self.ioctx):
            return
        self.rbd.create(self.ioctx, self.name, size)

    def ensure_size(self, size):
        if self.size == size:
            return
        self.image.resize(size)

    def lock(self):
        logger.debug('Assuming lock for %s', self.fullname)
        retry = 3
        while retry:
            try:
                self.image.lock_exclusive(self.ceph.CEPH_LOCK_HOST)
            except rbd.ImageExists:
                # This client and cookie already locked this. This is
                # definitely fine.
                return
            except rbd.ImageBusy:
                # Maybe the same client but different cookie. We're fine with
                # different cookies - ignore this. Must be same client, though.
                status = self.lock_status()
                if status is None:
                    # Someone had the lock but released it in between.
                    # Lets try again
                    retry -= 1
                    continue
                if status[1] == self.ceph.CEPH_LOCK_HOST:
                    # That's locked for us already. We just re-use
                    # the existing lock.
                    return
                # Its locked for some other client.
                logger.error(
                    'Failed assuming lock. Giving up. '
                    'Competing lock: {}'.format(status))
                raise
        raise rbd.ImageBusy(
            'Could not acquire lock - tried multiple times. '
            'Someone seems to be racing me.')

    def lock_status(self):
        """Return None if not locked and (client_id, lock_id) if it is."""
        try:
            lockers = self.image.list_lockers()
        except rbd.ImageNotFound:
            return None
        if not lockers:
            return
        # For some reasons list_lockers may just return an empty list instead
        # of a dict. Say what?
        lockers = lockers['lockers']
        if not len(lockers) == 1:
            raise NotImplementedError("I'm not prepared for shared locks")
        client_id, lock_id, addr = lockers[0]
        return client_id, lock_id

    def unlock(self, force=False):
        locked_by = self.lock_status()
        if not locked_by:
            return
        client_id, lock_id = locked_by
        if not force and lock_id != self.ceph.CEPH_LOCK_HOST:
            raise rbd.ImageBusy("Can not break lock for {} held by host {}."
                                .format(self.name, lock_id))
        self.image.break_lock(client_id, lock_id)

    def mkswap(self):
        self.map()
        try:
            cmd('mkswap -f -L "{}" "/dev/rbd/{}"'.format(
                self.label, self.fullname))
        finally:
            self.unmap()

    def mkfs(self):
        self.map()
        try:
            cmd('mkfs -F -q -m 1 -t ext4 -L "{}" "/dev/rbd/{}"'.format(
                self.label, self.fullname))
            time.sleep(0.1)
            cmd('tune2fs -e remount-ro "/dev/rbd/{}"'.format(self.fullname))
        finally:
            self.unmap()

    def seed_enc(self, data):
        self.map()
        try:
            target = tempfile.mkdtemp(prefix='/mnt/create-vm.')
            cmd('mount /dev/rbd/{} {}'.format(self.fullname, target))
            try:
                os.mkdir(p.join(target, 'fc-data'))
                with open(p.join(target, 'fc-data', 'enc.json'), 'w') as f:
                    f.write(json.dumps(data))
            finally:
                cmd('umount {}'.format(target))
                shutil.rmtree(target)
        finally:
            self.unmap()

    def map(self):
        if self.mapped:
            return
        cmd('rbd --id "{}" map "{}"'.format(
            self.ceph.CEPH_CLIENT, self.fullname))
        time.sleep(0.1)
        self.mapped = True

    def unmap(self):
        if not self.mapped:
            return
        cmd('rbd --id "{}" unmap "/dev/rbd/{}"'.format(
            self.ceph.CEPH_CLIENT, self.fullname))
        self.mapped = False


class Snapshots(object):

    def __init__(self, volume):
        self.volume = volume

    def create(self, name):
        cmd('rbd --id "{}" snap create "{}@{}"'.format(
            self.volume.ceph.CEPH_CLIENT, self.volume.fullname, name))
        logger.info('Created snapshot %s@%s', self.volume.fullname, name)

    def list(self):
        output = cmd(
            'rbd --id "{}" --format=json snap ls "{}"'.format(
                self.volume.ceph.CEPH_CLIENT, self.volume.fullname))
        return json.loads(output)

    def remove(self, name):
        cmd('rbd --id "{}" snap rm "{}@{}"'.format(
            self.volume.ceph.CEPH_CLIENT, self.volume.fullname, name))


class Ceph(object):

    # Attributes on this class can be overriden in a controlled fashion
    # from the sysconfig module. See __init__(). The defaults are here to
    # support testing.

    CEPH_CLUSTER = None
    CEPH_LOCK_HOST = None
    CEPH_CLIENT = 'admin'
    CREATE_VM = None
    SHRINK_VM = None

    def __init__(self, cfg):
        # Update configuration values from system or test config.
        self.__dict__.update(sysconfig.ceph)

        self.cfg = cfg
        self.ceph_conf = '/etc/ceph/{}.conf'.format(self.CEPH_CLUSTER)
        self.rados = None
        self.ioctx = None
        self.root = None
        self.swap = None
        self.tmp = None
        self.volumes = []

    def __enter__(self):
        # Not sure whether it makes sense that we configure the client ID
        # without 'client.': qemu doesn't want to see this, whereas the
        # Rados binding does ... :/
        self.rados = rados.Rados(
            conffile=self.ceph_conf,
            name='client.' + self.CEPH_CLIENT)
        self.rados.connect()

        pool = self.cfg['resource_group'].encode('ascii')
        self.ioctx = self.rados.open_ioctx(pool)

        volume_prefix = self.cfg['name'].encode('ascii')
        self.root = Volume(self, volume_prefix + '.root', 'root')
        self.swap = Volume(self, volume_prefix + '.swap', 'swap')
        self.tmp = Volume(self, volume_prefix + '.tmp', 'tmp')

        self.volumes = [self.root, self.swap, self.tmp]

    def __exit__(self, exc_value, exc_type, exc_tb):
        self.ioctx.close()
        self.rados.shutdown()

    def start(self):
        self.ensure_root_volume()
        self.ensure_tmp_volume()
        self.ensure_swap_volume()

    def stop(self):
        self.unlock()

    def shrink_root(self):
        # Note: we trust the called script to lock and unlock
        # the root image.
        target_size = self.cfg['disk'] * 1024 ** 3
        if self.root.size <= target_size:
            return
        try:
            cmd(self.SHRINK_VM.format(image=self.root.name, **self.cfg))
        except subprocess.CalledProcessError:
            raise RuntimeError(
                'unrecoverable error while shrinking root volume',
                self.cfg['name'])

    def ensure_root_volume(self):
        if not self.root.exists():
            cmd(self.CREATE_VM.format(**self.cfg))
        self.shrink_root()
        self.root.lock()

    def ensure_swap_volume(self):
        self.swap.ensure_presence(self.cfg['swap_size'])
        self.swap.lock()
        self.swap.ensure_size(self.cfg['swap_size'])
        self.swap.mkswap()

    def ensure_tmp_volume(self):
        self.tmp.ensure_presence(self.cfg['tmp_size'])
        self.tmp.lock()
        self.tmp.ensure_size(self.cfg['tmp_size'])
        self.tmp.mkfs()

    def locks(self):
        for vol in self.volumes:
            status = vol.lock_status()
            if not status:
                continue
            yield vol.name, status[1]

    def is_unlocked(self):
        """Returns True if no volume is locked."""
        return all(not vol.lock_status() for vol in self.volumes)

    def locked_by_me(self):
        """Returns True if CEPH_LOCK_HOST holds locks for all volumes."""
        try:
            return all(v.lock_status()[1] == self.CEPH_LOCK_HOST
                       for v in self.volumes)
        except TypeError:  # status[1] not accessible
            return False

    def lock(self):
        for vol in self.volumes:
            vol.lock()

    def unlock(self):
        """Remove all of *our* volume locks.

        This leaves other hosts' locks in place.
        """
        for vol in self.volumes:
            try:
                vol.unlock()
            except rbd.ImageBusy:
                pass

    def force_unlock(self):
        for vol in self.volumes:
            vol.unlock(force=True)

    def auth_cookie(self):
        """This is a cookie that can be used to validate that a party
        has access to Ceph.

        Used to authenticate migration requests.

        """
        c = hashlib.sha1()
        for vol in self.volumes:
            status = [vol.name]
            lock = vol.lock_status()
            if lock:
                status.extend(lock)
            c.update('\0'.join(status) + '\0')
        return c.hexdigest()



import ConfigParser
import os.path


class SysConfig(object):
    """A global config state registry.

    This is used to manage system-specific configuration overrides
    for values used in various places within fc.qemu.

    This provides the code for loading all those options from a central
    config file and to allow tests overriding those values gracefully.

    """

    def __init__(self):
        self.qemu = {}
        self.ceph = {}
        self.agent = {}

    def load_system_config(self):
        # System-wide config - pretty hacky
        sysconfig = ConfigParser.SafeConfigParser()
        sysconfig.read(os.path.dirname(__file__) + '/default.conf')
        sysconfig.read('/etc/qemu/fc-qemu.conf')

        self.qemu['migration_address'] = sysconfig.get(
            'qemu', 'migration-address')
        self.qemu['require_kvm'] = bool(sysconfig.get('qemu', 'accelerator'))
        self.qemu['vnc'] = sysconfig.get('qemu', 'vnc')
        self.qemu['max_downtime'] = sysconfig.getfloat('qemu', 'max-downtime')

        self.agent['accelerator'] = sysconfig.get('qemu', 'accelerator')
        self.agent['ceph_id'] = sysconfig.get('ceph', 'client-id')
        self.agent['consul_token'] = sysconfig.get('consul', 'access-token')
        self.agent['migration_ctl_address'] = sysconfig.get(
            'qemu', 'migration-ctl-address')
        self.agent['timeout_graceful'] = sysconfig.getint(
            'qemu', 'timeout-graceful')
        self.agent['this_host'] = sysconfig.get('ceph', 'lock_host')
        self.agent['vhost'] = sysconfig.getboolean('qemu', 'vhost')

        self.ceph['CEPH_CLIENT'] = sysconfig.get('ceph', 'client-id')
        self.ceph['CEPH_CLUSTER'] = sysconfig.get('ceph', 'cluster', 'ceph')
        self.ceph['CEPH_LOCK_HOST'] = sysconfig.get('ceph', 'lock_host')
        self.ceph['CREATE_VM'] = sysconfig.get('ceph', 'create-vm')
        self.ceph['SHRINK_VM'] = sysconfig.get('ceph', 'shrink-vm')


sysconfig = SysConfig()



from fc.qemu.exc import MigrationError
from fc.qemu.incoming import authenticated, parse_address
from fc.qemu.incoming import IncomingAPI, IncomingServer
import mock
import pytest


def test_authentication_wrapper():
    @authenticated
    def test(cookie):
        return 1

    context = mock.Mock()
    context.cookie = 'asdf'
    assert test(context, 'asdf') == 1

    with pytest.raises(MigrationError):
        test(context, 'foobar')


def test_parse_address_host():
    assert ('host', 1234) == parse_address('host:1234')


def test_parse_address_ipv4():
    assert ('10.1.2.3', 4567) == parse_address('10.1.2.3:4567')


def test_parse_address_ipv5():
    assert ('2001:db8:1::45', 1234) == parse_address('[2001:db8:1::45]:1234')


def test_incoming_api():
    server = mock.Mock()
    server.agent.ceph.auth_cookie.return_value = 'asdf'
    api = IncomingAPI(server)
    assert api.cookie == 'asdf'

    api.ping('asdf')
    assert server.extend_cutoff_time.call_args_list == [mock.call()]

    api.acquire_locks('asdf')
    assert server.acquire_locks.call_args_list == [mock.call()]

    api.prepare_incoming('asdf', [], {})
    assert server.prepare_incoming.call_args_list == [mock.call([], {})]

    api.finish_incoming('asdf')
    assert server.finish_incoming.call_args_list == [mock.call()]

    api.rescue('asdf')
    assert server.rescue.call_args_list == [mock.call()]

    api.destroy('asdf')
    assert server.destroy.call_args_list == [mock.call()]

    api.cancel('asdf')
    assert server.cancel.call_args_list == [mock.call()]


def test_incoming_server():
    agent = mock.Mock()
    agent.migration_ctl_address = 'localhost:9000'
    server = IncomingServer(agent)
    assert server.bind_address == ('localhost', 9000)

    server._now = mock.Mock(return_value=30)
    server.extend_cutoff_time()
    assert server.timeout.cutoff == 90



from .timeout import TimeOut
import logging
import socket
import xmlrpclib

_log = logging.getLogger(__name__)


class Outgoing(object):

    migration_exitcode = None
    target = None
    cookie = None

    def __init__(self, agent):
        self.agent = agent
        self.name = agent.name
        self.consul = agent.consul

    def __call__(self):
        self.cookie = self.agent.ceph.auth_cookie()
        with self:
            self.connect()
            self.transfer_locks()
            self.migrate()

        return self.migration_exitcode

    # The context manager is suitable to express "I am going to
    # interact with this VM: please prepare connections to
    # the according backends and make sure to clean up after me.""
    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        if _exc_type is None:
            self.migration_exitcode = 0
        else:
            self.migration_exitcode = 1
            try:
                _log.debug('A problem occured trying to migrate the VM. '
                           'Trying to rescue it.',
                           exc_info=(_exc_type, _exc_value, _traceback))
                self.rescue()
                return True  # swallow exception
            except:
                # Purposeful bare except: try really hard to kill
                # our VM.
                _log.exception('A problem occured trying to rescue the VM '
                               'after a migration failure. Destroying it.')
                self.destroy()

    def locate_inmigrate_service(self, timeout=330):
        service_name = 'vm-inmigrate-' + self.name
        timeout = TimeOut(timeout, interval=3, raise_on_timeout=True)
        while timeout.tick():
            _log.debug('%s: waiting (%ds remaining)', self.name,
                       int(timeout.remaining))
            inmig = self.consul.catalog.service(service_name)
            if inmig:
                if len(inmig) > 1:
                    _log.warning('found more than one inmig services for %s, '
                                 'trying the most recent one', service_name)
                inmig = inmig[-1]
                _log.debug('found incoming service %r', inmig)
                return 'http://{}:{}'.format(
                    inmig['ServiceAddress'], inmig['ServicePort'])
        raise RuntimeError('failed to locate inmigrate service', service_name)

    def connect(self, timeout=330):
        timeout = TimeOut(timeout, interval=3, raise_on_timeout=True)
        while timeout.tick():
            try:
                address = self.locate_inmigrate_service()
                _log.debug('connecting to {}'.format(address))
                self.target = xmlrpclib.ServerProxy(address, allow_none=True)
                self.target.ping(self.cookie)
                break
            # XXX the default socket timeout is quite high. we might wanna
            # lower it on this side via socket.settimeout()
            except socket.error:
                _log.debug(
                    'cannot establish XML-RPC connection, retrying for %ds',
                    timeout.remaining)

    def transfer_locks(self):
        self.agent.ceph.unlock()
        self.target.acquire_locks(self.cookie)

    def migrate(self):
        """Actually move VM between hosts."""
        args, config = self.agent.qemu.get_running_config()
        _log.info('%s: preparing remote environment', self.name)
        migration_address = self.target.prepare_incoming(
            self.cookie, args, config)
        _log.info('%s: starting transfer to %s', self.name, migration_address)
        self.agent.qemu.migrate(migration_address)
        # XXX The status polling is a bit fuzzy: we noticed an additional
        # intermediate status "setup" while running this in production.
        # Over the long term this should get more reliable towards unknown
        # states and tolerate them (until running in a timeout)
        for stat in self.agent.qemu.monitor.poll_migration_status(
                'Migration status: completed',
                ['Migration status: active',
                 'Migration status: setup']):
            self.target.ping(self.cookie)

        self.agent.qemu.monitor.assert_status(
            'VM status: paused (postmigrate)')
        _log.info('%s: finishing and cleaning up', self.name)
        self.target.finish_incoming(self.cookie)
        self.agent.qemu.destroy()

    def rescue(self):
        """Outgoing rescue: try to rescue the remote side first."""
        _log.exception('{}: something went wrong, trying to rescue'.format(self.name))
        if self.target is not None:
            try:
                self.target.rescue(self.cookie)
                _log.info('%s: remote rescue successful, destroying our '
                          'instance', self.name)
                self.destroy()
                return
            except Exception as e:
                _log.debug(e)
                try:
                    _log.info('%s: remote rescue failed, trying remote '
                              'destroy', self.name)
                    self.target.destroy(self.cookie)
                except Exception as e:
                    _log.debug(e)
                    _log.info('%s: failed to destroy remote VM', self.name)
        try:
            self.agent.ceph.lock()
            _log.info('%s: re-acquired Ceph locks successfully, continuing '
                      'VM locally', self.name)
        except Exception:
            _log.warning('%s: failed to (re-)aquire Ceph locks, bailing out',
                         self.name)
            self.destroy()

    def destroy(self):
        self.agent.qemu.destroy()
        self.agent.qemu.clean_run_files()



from ..ceph import Ceph, Volume
import rbd
import os.path
import pytest


@pytest.yield_fixture
def ceph_inst():
    cfg = {'resource_group': 'test', 'name': 'test00', 'disk': 10}
    ceph = Ceph(cfg)
    ceph.CREATE_VM = 'echo {name}'
    ceph.__enter__()
    yield ceph
    ceph.__exit__(None, None, None)


@pytest.yield_fixture
def volume(ceph_inst):
    volume = Volume(ceph_inst, 'othervolume', 'label')

    try:
        for snapshot in volume.snapshots.list():
            volume.snapshots.remove(snapshot['name'])
    except Exception:
        pass

    try:
        rbd.RBD().remove(ceph_inst.ioctx, 'othervolume')
    except rbd.ImageNotFound:
        pass

    yield volume

    lock = volume.lock_status()
    if lock is not None:
        volume.image.break_lock(*lock)
    for snapshot in volume.snapshots.list():
        volume.snapshots.remove(snapshot['name'])
    rbd.RBD().remove(ceph_inst.ioctx, 'othervolume')


@pytest.yield_fixture
def ceph_with_volumes(ceph_inst):
    for vol in ceph_inst.volumes:
        vol.ensure_presence()
        vol.lock()
    yield ceph_inst
    for vol in ceph_inst.volumes:
        vol.unlock(force=True)
        rbd.RBD().remove(ceph_inst.ioctx, vol.name)


def test_volume_presence(volume):
    assert volume.fullname == 'test/othervolume'
    assert not volume.exists()
    with pytest.raises(rbd.ImageNotFound):
        volume.image
    volume.ensure_presence()
    assert volume.image
    # Check that ensure_presence is fine with being called multiple times.
    volume.ensure_presence()


def test_volume_snapshot(volume):
    volume.ensure_presence()
    assert volume.image
    volume.snapshots.create('test')
    snaps = volume.snapshots.list()
    assert len(snaps) == 1
    assert snaps[0]['name'] == 'test'

    volume.snapshots.remove('test')
    assert volume.snapshots.list() == []


def test_volume_size(volume):
    volume.ensure_presence()
    assert volume.image
    assert volume.size == 1024
    volume.ensure_size(2048)
    assert volume.size == 2048
    # Call ensure multiple times to help triggering caching code paths.
    volume.ensure_size(2048)
    assert volume.size == 2048


def test_volume_shared_lock_protection(volume):
    volume.ensure_presence()
    volume.image.lock_shared('host1', 'a')
    volume.image.lock_shared('remotehost', 'a')
    with pytest.raises(NotImplementedError):
        volume.lock_status()
    lockers = volume.image.list_lockers()
    for client, cookie, _ in lockers['lockers']:
        volume.image.break_lock(client, cookie)


def test_volume_locking(volume):
    # Non-existing volumes report None as locking status but do not raise
    # an exception.
    assert not volume.exists()
    assert volume.lock_status() is None
    volume.ensure_presence()
    assert volume.lock_status() is None
    volume.lock()
    assert volume.lock_status()[1] == 'host1'
    # We want to smoothen out that some other process has locked the same image
    # for the same tag already and assume that this is another incarnation of
    # us - for that we have our own lock.
    volume.lock()
    assert volume.lock_status()[1] == 'host1'
    volume.unlock()
    assert volume.lock_status() is None
    # We can call unlock twice if it isn't locked.
    volume.unlock()

    volume.image.lock_exclusive('someotherhost')
    with pytest.raises(rbd.ImageBusy):
        volume.lock()
    with pytest.raises(rbd.ImageBusy):
        # Can not unlock locks that someone else holds.
        volume.unlock()


def test_force_unlock(volume):
    volume.ensure_presence()
    volume.image.lock_exclusive('someotherhost')
    volume.unlock(force=True)
    assert volume.lock_status() is None


def test_volume_mkswap(volume):
    volume.ensure_presence()
    volume.ensure_size(5 * 1024 ** 2)
    volume.mkswap()


def test_volume_mkfs(volume):
    volume.ensure_presence()
    volume.ensure_size(5 * 1024 ** 2)
    volume.mkfs()


def test_volume_map_unmap(volume):
    volume.ensure_presence()
    volume.map()
    assert os.path.exists('/dev/rbd/test/othervolume')
    volume.map()
    assert os.path.exists('/dev/rbd/test/othervolume')
    volume.unmap()
    assert not os.path.exists('/dev/rbd/test/othervolume')
    volume.unmap()
    assert not os.path.exists('/dev/rbd/test/othervolume')


def test_call_shrink_vm(ceph_inst, capfd):
    try:
        rbd.RBD().remove(ceph_inst.ioctx, 'test00.root')
    except rbd.ImageNotFound:
        pass

    ceph_inst.root.ensure_presence(ceph_inst.cfg['disk'] * 1024 ** 3 + 4096)
    try:
        ceph_inst.ensure_root_volume()
        stdout, stderr = capfd.readouterr()
        assert 'shrink-vm pool=test image=test00.root disk=10' in stdout
    finally:
        ceph_inst.root.unlock()
        rbd.RBD().remove(ceph_inst.ioctx, ceph_inst.root.name)


def test_shrink_vm_raises_if_nonzero_exit(ceph_inst):
    ceph_inst.root.ensure_presence(ceph_inst.cfg['disk'] * 1024 ** 3 + 4096)
    ceph_inst.SHRINK_VM = '/bin/false'
    try:
        with pytest.raises(RuntimeError):
            ceph_inst.shrink_root()
    finally:
        rbd.RBD().remove(ceph_inst.ioctx, ceph_inst.root.name)


def test_ceph_stop_should_unlock_all_volumes(ceph_with_volumes):
    for vol in ceph_with_volumes.volumes:
        assert vol.lock_status()
    ceph_with_volumes.stop()
    for vol in ceph_with_volumes.volumes:
        assert vol.lock_status() is None


def test_ceph_stop_remove_only_own_locks(ceph_with_volumes):
    """Test case where failed migrations leave inconsistent locking."""
    ceph_with_volumes.root.unlock()
    ceph_with_volumes.root.image.lock_exclusive('someotherhost')
    ceph_with_volumes.stop()
    assert ceph_with_volumes.root.lock_status()
    assert ceph_with_volumes.swap.lock_status() is None
    assert ceph_with_volumes.tmp.lock_status() is None



from ...exc import MigrationError
from ..monitor import Monitor
import pytest
import socket
import telnetlib
import time


class FakeTelnet(object):

    def __init__(self, host, port, timeout=0):
        pass

    def write(self, data):
        pass

    def close(self):
        pass


class MigrationStatusSequence(FakeTelnet):

    # This is a class-mutable by intention: we need several connections
    # and keep track over those.
    attempts = []

    def read_until(self, search, timeout=0):
        if len(self.attempts) < 2:
            self.attempts.append(search)
            return 'Migration status: active\r\n(qemu)\r\n'
        return 'Migration status: completed\r\n(qemu)\r\n'


class ConnectionRefusedTelnet(FakeTelnet):

    def __init__(self, host, port, timeout=0):
        raise socket.error(111, 'Connection refused')


class TestMonitor(object):

    def test_cmd_sanitizes_info_status(self, monkeypatch):
        """The telnet monitor interface returns screwed echo."""
        monkeypatch.setattr(telnetlib, 'Telnet', FakeTelnet)
        # this crazy stuff is what we actually get from KVM :-/
        setattr(FakeTelnet, 'read_until', lambda self, exp, timeout: (
            ' i\x1b[K\x1b[Din\x1b[K\x1b[D\x1b[Dinf\x1b[K\x1b[D\x1b[D\x1b[Dinfo'
            '\x1b[K\x1b[D\x1b[D\x1b[D\x1b[Dinfo \x1b[K\x1b[D\x1b[D\x1b[D\x1b[D'
            '\x1b[Dinfo s\x1b[K\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[Dinfo st'
            '\x1b[K\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[Dinfo sta\x1b[K'
            '\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[Dinfo stat\x1b[K'
            '\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[Dinfo statu'
            '\x1b[K\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D\x1b[D'
            '\x1b[Dinfo status\x1b[K\r\nVM status: running\r\n(qemu)'))
        m = Monitor(12345)
        assert 'VM status: running\n' == m._cmd('info status')

    def test_cmd_raises_unless_prompt_in_output(self, monkeypatch):
        monkeypatch.setattr(telnetlib, 'Telnet', FakeTelnet)
        setattr(FakeTelnet, 'read_until', lambda self, exp, timeout: '')
        m = Monitor(12345)
        with pytest.raises(RuntimeError):
            m._cmd('info status')

    def test_status_should_not_catch_connection_errors(self, monkeypatch):
        monkeypatch.setattr(telnetlib, 'Telnet', ConnectionRefusedTelnet)
        m = Monitor(12345)
        with pytest.raises(socket.error):
            m._cmd('info status')

    def test_migstatus_should_catch_connection_errors(self, monkeypatch):
        monkeypatch.setattr(telnetlib, 'Telnet', ConnectionRefusedTelnet)
        m = Monitor(12345)
        assert m.info_migrate() == ''

    def test_poll_status_should_terminate_on_reached_state(self, monkeypatch):
        MigrationStatusSequence.attempts = []
        monkeypatch.setattr(telnetlib, 'Telnet', MigrationStatusSequence)
        monkeypatch.setattr(time, 'sleep', lambda t: None)
        m = Monitor(12345)
        res = list(m.poll_migration_status('Migration status: completed', [
            'Migration status: active', 'Migration status: starting']))
        assert 'Migration status: completed' in res[-1]

    def test_poll_status_should_raise_on_unexpected_status(self, monkeypatch):
        MigrationStatusSequence.attempts = []
        monkeypatch.setattr(telnetlib, 'Telnet', MigrationStatusSequence)
        monkeypatch.setattr(time, 'sleep', lambda t: None)
        m = Monitor(12345)
        with pytest.raises(MigrationError):
            for s in m.poll_migration_status('Migration status: completed', [
                    'interim status 1', 'interim status 2']):
                pass



class LockError(RuntimeError):
    pass


class MigrationError(RuntimeError):
    pass


class DestructionError(RuntimeError):
    """Failed to destroy a running VM."""
    pass


class QemuNotRunning(Exception):
    """Something happened and we're sure Qemu isn't running."""
    pass



from ..hazmat.qemu import Qemu
import os
import psutil
import pytest
import tempfile


@pytest.yield_fixture
def qemu_with_pidfile():
    q = Qemu(dict(name='testvm', id=1234))
    tf = tempfile.NamedTemporaryFile(prefix='test.', delete=False)
    q.pidfile = tf.name
    tf.write('{}\n'.format(os.getpid()))
    tf.close()
    try:
        yield q
    finally:
        try:
            os.unlink(tf.name)
        except OSError:
            pass


def test_proc_running(qemu_with_pidfile):
    assert isinstance(qemu_with_pidfile.proc(), psutil.Process)


def test_proc_not_running(qemu_with_pidfile):
    with open(qemu_with_pidfile.pidfile, 'w') as p:
        p.write('0\n')
    assert qemu_with_pidfile.proc() is None


def test_proc_no_pidfile(qemu_with_pidfile):
    os.unlink(qemu_with_pidfile.pidfile)
    assert qemu_with_pidfile.proc() is None


def test_proc_empty_pidfile(qemu_with_pidfile):
    open(qemu_with_pidfile.pidfile, 'w').close()
    assert qemu_with_pidfile.proc() is None


def test_proc_pidfile_with_garbage(qemu_with_pidfile):
    with open(qemu_with_pidfile.pidfile, 'a') as f:
        f.write('trailing line\n')
    assert isinstance(qemu_with_pidfile.proc(), psutil.Process)



from .hazmat.ceph import Ceph
from .hazmat.qemu import Qemu
from .incoming import IncomingServer
from .outgoing import Outgoing
from .timeout import TimeOut
from .util import rewrite, locate_live_service
from .sysconfig import sysconfig
from logging import getLogger
import consulate
import contextlib
import copy
import fcntl
import json
import multiprocessing
import os
import os.path as p
import pkg_resources
import socket
import sys
import time
import yaml

log = getLogger(__name__)


def _handle_consul_event(event):
    handler = ConsulEventHandler()
    handler.handle(event)


class ConsulEventHandler(object):

    def handle(self, event):
        """Actual handling of a single Consul event in a
        separate process."""
        try:
            prefix = event['Key'].split('/')[0]
            if not event['Value']:
                return
            getattr(self, prefix)(event)
        except Exception as e:
            log.exception('error handling consul event: %s', e)

    def node(self, event):
        config = json.loads(event['Value'].decode('base64'))
        config['consul-generation'] = event['ModifyIndex']
        vm = config['name']
        if config['parameters']['machine'] != 'virtual':
            return
        try:
            agent = Agent(vm, config)
        except RuntimeError:
            log.warning('[%s] Ignoring VM check: cannot load config', vm)
            return
        log.info('[%s] Checking VM', vm)
        with agent:
            agent.save_enc()
            agent.ensure()

    def snapshot(self, event):
        value = json.loads(event['Value'].decode('base64'))
        vm = value['vm']
        snapshot = value['snapshot']
        try:
            agent = Agent(vm)
        except RuntimeError:
            log.warning('Ignoring snapshot for %s: failed to load config', vm)
            return
        with agent:
            if not agent.belongs_to_this_host():
                log.debug('Ignoring snapshot for %s as it belongs to another '
                          'host.', vm)
                return
            log.info('[%s] Processing snapshot request %s', vm, snapshot)
            agent.snapshot(snapshot)


def running(expected=True):
    def wrap(f):
        def checked(self, *args, **kw):
            if self.qemu.is_running() != expected:
                raise RuntimeError(
                    'action not allowed - VM must {}be running here'.format(
                        '' if expected else 'not '))
            return f(self, *args, **kw)
        return checked
    return wrap


def locked(f):
    def locked(self, *args, **kw):
        if not self._configfile_fd:
            self._configfile_fd = os.open(self.configfile, os.O_RDONLY)
        fcntl.flock(self._configfile_fd, fcntl.LOCK_EX)
        try:
            return f(self, *args, **kw)
        finally:
            fcntl.flock(self._configfile_fd, fcntl.LOCK_UN)
    return locked


def swap_size(memory):
    if memory > 2048:
        swap = memory / 2
    else:
        swap = 1024
    return swap * 1024 ** 2


def tmp_size(disk):
    # disk in GiB, return Bytes
    return max(5 * 1024, disk * 1024 / 10) * 1024 ** 2


class Agent(object):
    """The agent to control a single VM."""

    # Attributes on this class can be overriden (in a controlled fashion
    # from the sysconfig module. See this class' __init__. The defaults
    # are here to support testing.

    this_host = ''
    migration_ctl_address = None
    accelerator = ''
    vhost = False
    ceph_id = 'admin'
    timeout_graceful = 80
    vm_config_template_path = [
        '/etc/qemu/qemu.vm.cfg.in',
        pkg_resources.resource_filename(__name__, 'qemu.vm.cfg.in')
    ]
    consul_token = None

    _configfile_fd = None

    def __init__(self, name, enc=None):
        # Update configuration values from system or test config.
        self.__dict__.update(sysconfig.agent)

        if '.' in name:
            self.configfile = name
        else:
            self.configfile = '/etc/qemu/vm/{}.cfg'.format(name)
        if enc is not None:
            self.enc = enc
        else:
            self.enc = self._load_enc()

        self.cfg = copy.copy(self.enc['parameters'])
        self.name = self.enc['name']
        self.cfg['name'] = self.enc['name']
        self.cfg['swap_size'] = swap_size(self.cfg['memory'])
        self.cfg['tmp_size'] = tmp_size(self.cfg['disk'])
        self.cfg['ceph_id'] = self.ceph_id
        self.qemu = Qemu(self.cfg)
        self.ceph = Ceph(self.cfg)
        self.contexts = [self.qemu, self.ceph]
        for attr in ['migration_ctl_address']:
            setattr(self, attr, getattr(self, attr).format(**self.cfg))
        for cand in self.vm_config_template_path:
            if p.exists(cand):
                self.vm_config_template = cand
                break
        self.consul = consulate.Consul(token=self.consul_token)

    def _load_enc(self):
        try:
            with open(self.configfile) as f:
                return yaml.safe_load(f)
        except IOError:
            raise RuntimeError("Could not load {}".format(self.configfile))

    @classmethod
    def handle_consul_event(cls):
        events = json.load(sys.stdin)
        if not events:
            return
        log.info('Processing %d consul event(s)', len(events))
        for e in events:
            p = multiprocessing.Process(target=_handle_consul_event, args=(e,))
            p.start()
            time.sleep(0.1)
        for proc in multiprocessing.active_children():
            proc.join()
        log.info('Finished processing %d consul event(s)', len(events))

    def save_enc(self):
        if not p.isdir(p.dirname(self.configfile)):
            os.makedirs(p.dirname(self.configfile))
        with rewrite(self.configfile) as f:
            yaml.safe_dump(self.enc, f)
        os.chmod(self.configfile, 0o644)

    def __enter__(self):
        for c in self.contexts:
            c.__enter__()

    def __exit__(self, exc_value, exc_type, exc_tb):
        for c in self.contexts:
            try:
                c.__exit__(exc_value, exc_type, exc_tb)
            except Exception:
                log.exception('Error while leaving agent contexts.')

    def belongs_to_this_host(self):
        return self.cfg['kvm_host'] == self.this_host

    @locked
    def ensure(self):
        if not self.cfg['online'] or not self.cfg['kvm_host']:
            self.ensure_offline()
        elif not self.belongs_to_this_host():
            if self.qemu.is_running():
                self.outmigrate()
            else:
                pass
        else:
            self.ensure_online()
            self.ensure_online_disk_size()

        if self.state_is_consistent():
            if not self.qemu.is_running():
                self.qemu.clean_run_files()
        else:
            log.warning('%s: state not consistent (monitor, pidfile, ceph), '
                        'destroying VM and cleaning up', self.name)
            self.qemu.destroy()

    def ensure_offline(self):
        if not self.qemu.is_running():
            return
        log.info('VM %s should not be running here', self.name)
        self.stop()

    def ensure_online(self):
        if self.qemu.is_running():
            # re-register in case services got lost during Consul restart
            self.consul_register()
            return
        log.info('VM %s should be running here', self.name)
        existing = locate_live_service(self.consul, 'qemu-' + self.name)
        if existing and existing['Address'] != self.this_host:
            log.info('Found VM %s to be running on %s already. '
                     'Trying an inmigration.', self.name, existing['Address'])
            self.inmigrate()
        else:
            self.start()

    def ensure_online_disk_size(self):
        """Trigger block resize action for the root disk via Qemu monitor."""
        target_size = self.cfg['disk'] * (1024 ** 3)
        if self.ceph.root.image.size() >= target_size:
            return
        log.info('Online disk resize for VM %s to %s GiB', self.name,
                 self.cfg['disk'])
        self.qemu.resize_root(target_size)

    def consul_register(self):
        """Register running VM with Consul."""
        self.consul.agent.service.register(
            'qemu-{}'.format(self.name),
            address=self.this_host,
            interval='5s',
            check=('test -e /proc/$(< /run/qemu.{}.pid )/mem || exit 2'.
                   format(self.name)))

    def consul_deregister(self):
        """De-register non-running VM with Consul."""
        self.consul.agent.service.deregister('qemu-{}'.format(self.name))

    @locked
    @running(False)
    def start(self):
        self.generate_config()
        log.info('Using Qemu config template %s', self.vm_config_template)
        self.ceph.start()
        self.ceph.tmp.seed_enc(self.enc)
        log.info('Starting VM %s', self.name)
        self.qemu.start()
        self.consul_register()
        # We exit here without releasing the ceph lock in error cases
        # because the start may have failed because of an already running
        # process. Removing the lock in that case is dangerous. OTOH leaving
        # the lock is never dangerous. Lets go with that.

    @contextlib.contextmanager
    def frozen_vm(self):
        try:
            try:
                log.debug('[%s] Freezing root disk.', self.name)
                self.qemu.freeze()
            except socket.timeout:
                log.warning('[%s] Timed out freezing the machine. '
                            'Continuing with unclean snapshot.', self.name)
            yield
        finally:
            try:
                log.debug('[%s] Thawing root disk.', self.name)
                self.qemu.thaw()
            except socket.timeout:
                log.warning('[%s] Failed to thaw. Retrying.', self.name)
                self.qemu.thaw()

    @locked
    def snapshot(self, snapshot):
        if snapshot in [x['name'] for x in self.ceph.root.snapshots.list()]:
            return
        if self.qemu.is_running():
            with self.frozen_vm():
                self.ceph.root.snapshots.create(snapshot)
        else:
            log.info('[%s] VM not running, creating snapshot without freezing',
                     self.name)
            self.ceph.root.snapshots.create(snapshot)

    def status(self):
        """Determine status of the VM."""
        if self.qemu.is_running():
            status = 0
            print('online')
        else:
            status = 1
            print('offline')
        for lock in self.ceph.locks():
            print('lock: {}@{}'.format(*lock))
        return status

    @locked
    def stop(self):
        timeout = TimeOut(self.timeout_graceful, interval=3)
        log.info('Trying graceful shutdown of VM %s...', self.name)
        try:
            self.qemu.graceful_shutdown()
        except (socket.error, RuntimeError):
            pass
        while timeout.tick():
            if not self.qemu.is_running():
                self.ceph.stop()
                self.qemu.clean_run_files()
                self.consul_deregister()
                log.info('Graceful shutdown of %s succeeded', self.name)
                break
        else:
            self.kill()

    @locked
    def restart(self):
        log.info('Restarting VM %s', self.name)
        self.stop()
        self.start()

    @locked
    @running(True)
    def kill(self):
        log.info('Killing VM %s', self.name)
        timeout = TimeOut(15, interval=1, raise_on_timeout=True)
        self.qemu.destroy()
        while timeout.tick():
            if not self.qemu.is_running():
                log.info('VM killed, cleaning up.')
                self.ceph.stop()
                self.qemu.clean_run_files()
                self.consul_deregister()
                break
        else:
            log.warning('Did not see the VM disappear. '
                        'Please check lock consistency.')

    @locked
    @running(False)
    def inmigrate(self):
        log.info('Preparing to migrate-in VM %s', self.name)
        if self.ceph.is_unlocked():
            log.info("%s: VM isn't running at all, starting it directly",
                     self.name)
            self.start()
            return
        server = IncomingServer(self)
        exitcode = server.run()
        if not exitcode:
            self.consul_register()
        log.info('%s: inmigration finished with exitcode %s', self.name,
                 exitcode)
        return exitcode

    @locked
    @running(True)
    def outmigrate(self):
        log.info('Migrating VM %s out', self.name)
        # re-register in case services got lost during Consul restart
        self.consul_register()
        client = Outgoing(self)
        exitcode = client()
        if not exitcode:
            self.consul_deregister()
        log.info('Out-migration of VM %s finished with exitcode %s',
                 self.name, exitcode)
        return exitcode

    @locked
    def lock(self):
        log.info('Assuming all Ceph locks for VM %s', self.name)
        for vol in self.ceph.volumes:
            vol.lock()

    @locked
    @running(False)
    def unlock(self):
        log.info('Releasing all Ceph locks for VM %s', self.name)
        self.ceph.stop()

    @running(False)
    def force_unlock(self):
        log.info('Breaking all Ceph locks for VM %s', self.name)
        self.ceph.force_unlock()

    # Helper methods
    def state_is_consistent(self):
        """Returns True if all relevant components agree about VM state.

        If False, results from Qemu monitor, pidfile or Ceph differ.
        """
        substates = [
            self.qemu.is_running(),
            bool(self.qemu.proc()),
            self.ceph.locked_by_me()]
        log.info('[%s] current state: qemu=%s, proc=%s, locked=%s',
                 *([self.name] + substates))
        return any(substates) == all(substates)

    # CAREFUL: changing anything in this config files will cause maintenance w/
    # reboot of all VMs in the infrastructure.
    def generate_config(self):
        """Generate a new Qemu config (and options) for a freshly
        starting VM.

        The configs are intended to be non-host-specific.

        This two-step behaviour is needed to support migrating VMs
        and keeping arguments consistent while allowing to localize arguments
        to the host running or receiving the VM.

        """
        self.qemu.args = [
            '-daemonize',
            '-nodefaults',
            '-name {name},process=kvm.{name}',
            '-chroot {{chroot}}',
            '-runas nobody',
            '-serial file:/var/log/vm/{name}.log',
            '-display vnc={{vnc}}',
            '-pidfile {{pidfile}}',
            '-vga std',
            '-m {memory}',
            '-watchdog i6300esb',
            '-watchdog-action reset',
            '-readconfig {{configfile}}']
        self.qemu.args = [a.format(**self.cfg)
                          for a in self.qemu.args]

        vhost = '  vhost = "on"' if self.vhost else ''

        netconfig = []
        for net, net_config in sorted(self.cfg['interfaces'].items()):
            ifname = 't{}{}'.format(net, self.cfg['id'])
            netconfig.append("""
[device]
  driver = "virtio-net-pci"
  netdev = "{ifname}"
  mac = "{mac}"

[netdev "{ifname}"]
  type = "tap"
  ifname = "{ifname}"
  script = "/etc/kvm/kvm-ifup"
  downscript = "/etc/kvm/kvm-ifdown"
{vhost}
""".format(ifname=ifname, mac=net_config['mac'], vhost=vhost))

        with open(self.vm_config_template) as f:
            tpl = f.read()
        accelerator = (' accel = "{}"'.format(self.accelerator)
                       if self.accelerator else '')
        self.qemu.config = tpl.format(
            accelerator=accelerator,
            network=''.join(netconfig), **self.cfg)



import json
import logging
import random
import socket

log = logging.getLogger(__name__)


class ClientError(RuntimeError):
    pass


class GuestAgent(object):
    """Wraps qemu guest agent wire protocol."""

    def __init__(self, machine, timeout=5):
        self.machine = machine
        self.file = None
        self.client = None
        self.timeout = timeout

    def read(self):
        """Reads single response from the GA and returns the result.

        Blocks and runs into a timeout if no response is available.
        """
        result = json.loads(self.file.readline())
        if 'error' in result:
                raise ClientError(result)
        return result['return']

    def cmd(self, cmd, **args):
        """Issues GA command and returns the result."""
        self.client.send(json.dumps({"execute": cmd, "arguments": args}))
        return self.read()

    def sync(self):
        """Ensures that request and response are in order."""
        sync_id = random.randint(0, 0xffff)
        n = 0
        result = self.cmd('guest-sync', id=sync_id)
        while n < 3:
            if result == sync_id:
                return
            log.error('[%s] Received incorrect sync id. Expected: %s Got: %s',
                      self.machine, sync_id, result)
            n += 1
            result = self.read()
        raise ClientError('Unable to sync with guest agent after {} tries.'.
                          format(n))

    def __enter__(self):
        self.client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.client.settimeout(self.timeout)
        self.client.connect('/run/qemu.{}.gqa.sock'.format(self.machine))
        self.file = self.client.makefile()
        self.sync()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()
        self.client.close()



from ..util import rewrite
import os
import pytest
import tempfile


@pytest.yield_fixture
def tmp():
    tf = tempfile.NamedTemporaryFile(mode='w', delete=False)
    yield tf
    try:
        os.unlink(tf.name)
    except OSError:  # pragma: nocover
        pass


def test_rewrite(tmp):
    tmp.write('old content\n')
    tmp.flush()
    old_ino = os.fstat(tmp.fileno()).st_ino
    with rewrite(tmp.name) as f:
        f.write('new content\n')
    with open(tmp.name) as f:
        assert 'new content\n' == f.read()
    assert old_ino != os.stat(tmp.name).st_ino


def test_rewrite_should_ignore_idempotent_rewrite(tmp):
    tmp.write('old content\n')
    tmp.flush()
    old_ino = os.fstat(tmp.fileno()).st_ino
    with rewrite(tmp.name) as f:
        f.write('old content\n')
    assert old_ino == os.stat(tmp.name).st_ino


def test_rewrite_ignore_deleted(tmp):
    tmp.write('old content\n')
    tmp.flush()
    old_ino = os.fstat(tmp.fileno()).st_ino
    with rewrite(tmp.name) as f:
        os.unlink(f.name)
    with open(tmp.name) as f:
        assert 'old content\n' == f.read()
    assert old_ino == os.stat(tmp.name).st_ino


def test_rewrite_should_create_nonexisting_file(tmp):
    os.unlink(tmp.name)
    with rewrite(tmp.name) as f:
        f.write('new file\n')
    with open(tmp.name) as f:
        assert 'new file\n' == f.read()



"""Low-level interface to Qemu commands."""

from ..exc import QemuNotRunning
from ..sysconfig import sysconfig
from ..timeout import TimeOut
from .guestagent import GuestAgent, ClientError
from .monitor import Monitor
import glob
import logging
import os.path
import psutil
import socket
import subprocess
import time
import yaml


log = logging.getLogger(__name__)


class Qemu(object):

    executable = 'qemu-system-x86_64'

    # Attributes on this class can be overriden (in a controlled fashion
    # from the sysconfig module. See this class' __init__. The defaults
    # are here to support testing.

    cfg = None
    require_kvm = True
    migration_address = None
    max_downtime = 1.0

    # The non-hosts-specific config configuration of this Qemu instance.
    args = ()
    config = ''

    # Host-specific qemu configuration
    chroot = '/srv/vm/{name}'
    vnc = 'localhost:1'

    MONITOR_OFFSET = 20000

    pidfile = '/run/qemu.{name}.pid'
    configfile = '/run/qemu.{name}.cfg'
    argfile = '/run/qemu.{name}.args'

    def __init__(self, vm_cfg):
        # Update configuration values from system or test config.
        self.__dict__.update(sysconfig.qemu)

        self.cfg = vm_cfg
        # expand template keywords in configuration variables
        for f in ['pidfile', 'configfile', 'argfile', 'migration_address']:
            setattr(self, f, getattr(self, f).format(**vm_cfg))
        # We are running qemu with chroot which causes us to not be able to
        # resolve names. :( See #13837.
        a = self.migration_address.split(':')
        if a[0] == 'tcp':
            a[1] = socket.gethostbyname(a[1])
        self.migration_address = ':'.join(a)
        self.name = self.cfg['name']
        self.monitor_port = self.cfg['id'] + self.MONITOR_OFFSET
        self.guestagent = GuestAgent(self.name, timeout=30)

    def __enter__(self):
        self.monitor = Monitor(self.monitor_port)

    def __exit__(self, exc_value, exc_type, exc_tb):
        self.monitor = None

    def proc(self):
        """Qemu processes as psutil.Process object.

        Returns None if the PID file does not exist or the process is
        not running.
        """
        try:
            with open(self.pidfile) as p:
                # pid file may contain trailing lines with garbage
                for line in p:
                    return psutil.Process(int(line))
        except (IOError, OSError, ValueError, psutil.NoSuchProcess):
            pass

    def _start(self, additional_args=()):
        if self.require_kvm and not os.path.exists('/dev/kvm'):
            raise RuntimeError('Refusing to start without /dev/kvm support.')
        self.prepare_config()
        with open('/proc/sys/vm/compact_memory', 'w') as f:
            f.write('1\n')
        try:
            cmd = '{} {} {}'.format(
                self.executable,
                ' '.join(self.local_args),
                ' '.join(additional_args))
            # We explicitly close all fds for the child to avoid
            # inheriting locks infinitely.
            log.info('[qemu] %s: %s', self.name, cmd)
            subprocess.check_call(cmd, shell=True, close_fds=True)
        except subprocess.CalledProcessError:
            # Did not start. Not running.
            log.exception('[qemu] %s: Failed to start', self.name)
            raise QemuNotRunning()

    def start(self):
        self._start()
        assert self.is_running()

    def freeze(self):
        with self.guestagent as guest:
            try:
                guest.cmd('guest-fsfreeze-freeze')
            except ClientError:
                pass
            assert guest.cmd('guest-fsfreeze-status') == 'frozen'

    def thaw(self):
        with self.guestagent as guest:
            try:
                guest.cmd('guest-fsfreeze-thaw')
            except ClientError:
                pass
            assert guest.cmd('guest-fsfreeze-status') == 'thawed'

    def inmigrate(self):
        self._start(['-incoming {}'.format(self.migration_address)])
        time.sleep(1)
        self.monitor.assert_status('VM status: paused (inmigrate)')
        return self.migration_address

    def migrate(self, address):
        """Initiate actual (out-)migration"""
        log.debug('[qemu] %s: migrate (mon:%s)', self.name, self.monitor_port)
        self.monitor.migrate(address, self.max_downtime)

    def is_running(self):
        # This method must be very reliable. It is perfectly OK to error
        # out in the case of inconsistencies. But a "true" must mean:
        # we have a working Qemu instance here. And a "false" must mean:
        # there is no reason to think that any remainder of a Qemu process is
        # still running

        # a) is there a process?
        proc = self.proc()
        if proc is None:
            expected_process_exists = False
        else:
            expected_process_exists = proc.is_running()

        # b) is the monitor port around?
        monitor_exists = self.monitor.peek()

        # c) is the monitor available and talks to us?
        monitor_says_running = False
        status = ''
        if monitor_exists:
            timeout = TimeOut(10, 1, raise_on_timeout=False)
            while timeout.tick():
                try:
                    status = self.monitor.status()
                    monitor_says_running = status == 'VM status: running'
                except Exception as e:
                    log.exception('waiting for monitor status: %s', e)
                    pass
                else:
                    break

        if (expected_process_exists and
                monitor_exists and monitor_says_running):
            return True

        if (not expected_process_exists and not monitor_exists):
            # This is still a bit icky: there were cases where we did
            # not see the process, the monitor port was open,
            return False

        # Any other combination is sufficiently weird so as to give up right
        # now.
        raise RuntimeError(
            'Can not determine whether Qemu is running. '
            'Process: {} Monitor Port: {} '
            'Monitor output: {} Status: {}'.format(
                expected_process_exists, monitor_exists, monitor_says_running,
                status))

    def status(self):
        return self.monitor.status()

    def rescue(self):
        """Recover from potentially inconsistent state.

        If the VM is running and we own all locks, then everything is fine.

        If the VM is running and we do not own the locks, then try to acquire
        them or bail out.

        Returns True if we were able to rescue the VM.
        Returns False if the rescue attempt failed and the VM is stopped now.

        """
        # XXX hold a lock to avoid another process on the same machine to
        # interfere with the VM while we're on it. E.g. by locking the main
        # config file.
        self.monitor.assert_status('VM status: running')
        for image in set(self.locks.available) - set(self.locks.held):
            try:
                self.acquire_lock(image)
            except Exception:
                pass
        self.assert_locks()

    def graceful_shutdown(self):
        self.monitor.sendkey('ctrl-alt-delete')

    def destroy(self):
        # We use this destroy command in "fire-and-forget"-style because
        # sometimes the init script will complain even if we achieve what
        # we want: that the VM isn't running any longer. We check this
        # by contacting the monitor instead.
        timeout = TimeOut(100, interval=1, raise_on_timeout=True)
        while timeout.tick():
            p = self.proc()
            if p:
                p.terminate()
            if not self.monitor.peek():
                break

    def resize_root(self, size):
        size = size / 1024 ** 2  # MiB
        self.monitor._cmd('block_resize virtio0 {}'.format(size))

    def clean_run_files(self):
        log.debug('Removing run files')
        for runfile in glob.glob('/run/qemu.{}.*'.format(self.cfg['name'])):
            os.unlink(runfile)

    def prepare_config(self):
        if not os.path.exists('/var/log/vm'):
            os.makedirs('/var/log/vm')

        chroot = self.chroot.format(**self.cfg)
        if not os.path.exists(chroot):
            os.makedirs(chroot)

        format = lambda s: s.format(
            pidfile=self.pidfile,
            configfile=self.configfile,
            monitor_port=self.monitor.port,
            vnc=self.vnc.format(**self.cfg),
            chroot=chroot,
            **self.cfg)
        self.local_args = [format(a) for a in self.args]
        self.local_config = format(self.config)

        with open(self.configfile+'.in', 'w') as f:
            f.write(self.config)
        with open(self.configfile, 'w') as f:
            f.write(self.local_config)
        with open(self.argfile+'.in', 'w') as f:
            yaml.safe_dump(self.args, f)

        # Qemu tends to overwrite the pid file incompletely -> truncate
        open(self.pidfile, 'w').close()

    def get_running_config(self):
        """Return the host-independent version of the current running
        config."""
        with open(self.argfile + '.in') as a:
            args = yaml.safe_load(a.read())
        with open(self.configfile + '.in') as c:
            config = c.read()
        return args, config



"""Low-level Qemu Monitor abstraction."""

from ..exc import MigrationError
from fc.qemu.timeout import TimeOut
import logging
import re
import socket
import telnetlib


_log = logging.getLogger(__name__)


class Monitor(object):
    """KVM monitor connection.

    Provides easy to use abstraction for monitor commands.
    """

    def __init__(self, port, timeout=10):
        self.port = port
        self.timeout = timeout

    def _connect(self):
        conn = telnetlib.Telnet('localhost', self.port, self.timeout)
        res = conn.read_until('(qemu)', self.timeout)
        if '(qemu)' not in res:
            raise RuntimeError('failed to establish monitor connection', res)
        return conn

    def peek(self):
        """Checks whether the monitor port is TCP reachable."""
        addr = socket.getaddrinfo(
            'localhost', self.port, socket.AF_UNSPEC, socket.SOCK_STREAM, 0)
        for (af, socktype, proto, _cname, sockaddr) in addr:
            try:
                s = socket.socket(af, socktype, proto)
                s.settimeout(1)
                s.connect(sockaddr)
                return True
            except socket.error:
                continue
        return False

    def _cmd(self, command):
        """Issues a monitor command and return QEMU's response.

        The monitor connection will be established on the first
        invocation.
        """
        conn = self._connect()
        try:
            _log.debug('[mon:%s] %s', self.port, command)
            conn.write(command + '\n')
            res = conn.read_until('(qemu)', self.timeout)
            r_strip_echo = re.compile(r'^.*' + re.escape(command) + '\\S*\r\n')
            output = r_strip_echo.sub('', res).replace('\r\n', '\n')
            output = output.replace('(qemu)', '')
            _log.debug('[mon:%s] %s', self.port, output.strip())
            return output
        finally:
            conn.close()

    def status(self):
        """VM status summary.

        Returns one-line status string or an empty string if we cannot
        connect to the monitor.

        """
        status = self._cmd('info status').strip()
        if 'VM status:' not in status:
            raise AssertionError(
                'Got unexpected status output: {}'.format(status))
        return status

    def assert_status(self, expected):
        status = self.status()
        if status != expected:
            raise RuntimeError(
                'VM status mismatch: expected "{}" got "{}"'.format(
                    expected, status))

    def sendkey(self, keys):
        self._cmd('sendkey {}'.format(keys))

    def migrate(self, address, downtime):
        """Initiate migration (asynchronously)."""
        self._cmd('migrate_set_capability xbzrle on')
        self._cmd('migrate_set_capability auto-converge on')
        self._cmd('migrate_set_downtime {}'.format(downtime))
        res = self._cmd('migrate -d {}'.format(address)).strip()
        if res:
            raise MigrationError('error while initiating migration', res)

    def info_migrate(self):
        """Migration status and statistics."""
        try:
            return self._cmd('info migrate')
        except Exception:
            return ''

    def poll_migration_status(self, target, acceptable_interim, timeout=30):
        """Monitor ongoing migration.

        Every few seconds, the migration status is queried from the KVM
        monitor. It is yielded to the calling context to provide a hook
        for communicating status updates.

        The migration status is allowed to progress through any value of
        `acceptable_interim`. This function terminates when the status
        reaches `target`. If any status that is not in
        `acceptable_interim` nor `target` is reached, this function
        raises an exception.
        """
        timeout = TimeOut(timeout, 1, raise_on_timeout=True)
        startup_phase = True
        while timeout.tick():
            if timeout.interval < 10:
                timeout.interval *= 1.4142
            status = self.info_migrate()
            yield status
            if startup_phase and not status.strip():
                # The monitor didn't really respond with anything.
                # This tends to happen sometimes at the beginning of the
                # migration. I let this slip.
                continue
            if target in status:
                break
            if not any(i in status for i in acceptable_interim):
                raise MigrationError('invalid migration status', status)
            startup_phase = False
            timeout.cutoff += 30

    def quit(self):
        """Terminate KVM process."""
        try:
            self._cmd('quit')
        except (EOFError, MigrationError):
            pass
        else:
            raise RuntimeError('Machine did not quit?')



from .agent import Agent
from .sysconfig import sysconfig
import argparse
import logging
import os.path
import sys


logger = logging.getLogger(__name__)


def daemonize():
    """
    Copyright/License abberation:

    Copied from
http://stackoverflow.com/questions/1417631/python-code-to-daemonize-a-process
    and
http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/

    do the UNIX double-fork magic, see Stevens' "Advanced
    Programming in the UNIX Environment" for details (ISBN 0201563177)
    http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
    """
    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write(
            "fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

    # decouple from parent environment
    os.chdir("/")
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write(
            "fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

    # redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    si = open('/dev/null', 'r')
    so = open('/dev/null', 'a+', 0)
    se = open('/dev/null', 'a+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


def init_logging(verbose=True):
    if verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(
        filename='/var/log/fc-qemu.log',
        format='%(asctime)s [%(process)d] %(levelname)s %(message)s',
        level=level)

    # silence requests package -- we assume that it's just doing its job
    logging.getLogger('requests').setLevel(logging.CRITICAL)

    console = logging.StreamHandler(sys.stdout)
    console.setLevel(level)
    logging.getLogger('').addHandler(console)


def main():
    a = argparse.ArgumentParser(description="Qemu VM agent")
    a.add_argument('--verbose', '-v', action='store_true', default=False,
                   help='Increase logging level.')
    a.add_argument('--daemonize', '-D', action='store_true',
                   help="Run command in background.")

    sub = a.add_subparsers(title='subcommands')
    p = sub.add_parser('status', help='Get the status of a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='status')

    p = sub.add_parser('ensure', help='Ensure proper status of the VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='ensure')

    p = sub.add_parser('start', help='Start a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='start')

    p = sub.add_parser('stop', help='Stop a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='stop')

    p = sub.add_parser('restart', help='Restart a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='restart')

    p = sub.add_parser('lock', help='Assume all locks of a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='lock')

    p = sub.add_parser('unlock', help='Release all locks of a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='unlock')

    p = sub.add_parser('snapshot',
                       help='Take a clean snapshot of this VM\'s root volume.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.add_argument('snapshot', help='name of the snapshot')
    p.set_defaults(func='snapshot')

    p = sub.add_parser('force-unlock', help="Release all locks of a VM even "
                       "if we don't own them.")
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='force_unlock')

    p = sub.add_parser('inmigrate', help='Start incoming migration for a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='inmigrate')

    p = sub.add_parser('outmigrate', help='Start outgoing migration for a VM.')
    p.add_argument('vm', metavar='VM', help='name of the VM')
    p.set_defaults(func='outmigrate')

    p = sub.add_parser(
        'handle-consul-event',
        help='Handle a change in VM config distributed via consul.')
    p.set_defaults(func='handle_consul_event')

    args = a.parse_args()
    func = args.func
    vm = getattr(args, 'vm', None)
    kwargs = dict(args._get_kwargs())
    del kwargs['func']
    if 'vm' in kwargs:
        del kwargs['vm']
    del kwargs['daemonize']
    del kwargs['verbose']

    if args.daemonize:
        daemonize()

    try:
        init_logging(args.verbose)
        sysconfig.load_system_config()
        if vm is None:
            # Expecting a class/static method
            agent = Agent
            sys.exit(getattr(agent, func)(**kwargs) or 0)
        else:
            agent = Agent(vm)
            with agent:
                sys.exit(getattr(agent, func)(**kwargs) or 0)
    except Exception as e:
        logger.exception(e)
        sys.exit(69)  # EX_UNAVAILABLE



from ..agent import Agent
import mock
import os
import pkg_resources
import psutil
import pytest
import shutil


@pytest.yield_fixture
def simplevm_cfg():
    fixtures = pkg_resources.resource_filename(__name__, 'fixtures')
    shutil.copy(fixtures + '/simplevm.yaml', '/etc/qemu/vm/simplevm.cfg')
    yield 'simplevm'
    os.unlink('/etc/qemu/vm/simplevm.cfg')


def test_builtin_config_template(simplevm_cfg):
    a = Agent(simplevm_cfg)
    a.generate_config()
    # machine type must match Qemu version in virtualbox
    assert 'type = "pc-q35-2.1"' in a.qemu.config


def test_userdefined_config_template(simplevm_cfg):
    with open('/etc/qemu/qemu.vm.cfg.in', 'w') as f:
        f.write('# user defined config template\n')
    try:
        a = Agent(simplevm_cfg)
        a.generate_config()
        assert 'user defined config template' in a.qemu.config
    finally:
        os.unlink('/etc/qemu/qemu.vm.cfg.in')


def test_consistency_vm_running(simplevm_cfg):
    a = Agent(simplevm_cfg)
    a.qemu.is_running = mock.Mock(return_value=True)
    a.qemu.proc = mock.Mock(return_value=psutil.Process(1))
    a.ceph.locked_by_me = mock.Mock(return_value=True)
    assert a.state_is_consistent() is True


def test_consistency_vm_not_running(simplevm_cfg):
    a = Agent(simplevm_cfg)
    a.qemu.is_running = mock.Mock(return_value=False)
    a.qemu.proc = mock.Mock(return_value=None)
    a.ceph.locked_by_me = mock.Mock(return_value=False)
    assert a.state_is_consistent() is True


def test_consistency_process_dead(simplevm_cfg):
    a = Agent(simplevm_cfg)
    a.qemu.is_running = mock.Mock(return_value=True)
    a.qemu.proc = mock.Mock(return_value=None)
    a.ceph.locked_by_me = mock.Mock(return_value=True)
    assert a.state_is_consistent() is False


def test_consistency_pidfile_missing(simplevm_cfg):
    a = Agent(simplevm_cfg)
    a.qemu.is_running = mock.Mock(return_value=True)
    a.qemu.proc = mock.Mock(return_value=None)
    a.ceph.locked_by_me = mock.Mock(return_value=True)
    assert a.state_is_consistent() is False


def test_consistency_ceph_lock_missing(simplevm_cfg):
    a = Agent(simplevm_cfg)
    a.qemu.is_running = mock.Mock(return_value=True)
    a.qemu.proc = mock.Mock(return_value=psutil.Process(1))
    a.ceph.locked_by_me = mock.Mock(return_value=False)
    assert a.state_is_consistent() is False



import contextlib
import filecmp
import os
import tempfile


@contextlib.contextmanager
def rewrite(filename):
    """Rewrite an existing file atomically.

    Clients are allowed to delete the tmpfile to signal that they don't
    want to have it updated.
    """

    with tempfile.NamedTemporaryFile(
            dir=os.path.dirname(filename), prefix=os.path.basename(filename),
            delete=False) as tf:
        if os.path.exists(filename):
            os.chmod(tf.name, os.stat(filename).st_mode & 0o7777)
        yield tf
        if not os.path.exists(tf.name):
            return
        filename_tmp = tf.name
    if (os.path.exists(filename) and
            filecmp.cmp(filename, filename_tmp, shallow=False)):
        os.unlink(filename_tmp)
    else:
        os.rename(filename_tmp, filename)


def parse_address(addr):
    if addr.startswith('['):
        host, port = addr[1:].split(']:')
    else:
        host, port = addr.split(':')
    return host, int(port)


def locate_live_service(consul, service_id):
    """Locate Consul service with at least one passing health check.

    It is an error if multiple live services with the same service name
    are found.
    """
    passing = lambda checks: (
        any(check['Status'] == 'passing' for check in checks) and
        not any(check['Status'] == 'critical' for check in checks))
    live = [svc for svc in consul.health.service(service_id)
            if passing(svc['Checks'])]
    if len(live) > 1:
        raise RuntimeError('multiple services with passing checks found',
                           service_id)
    return live[0]['Service'] if len(live) else None



from .exc import MigrationError, QemuNotRunning
from .timeout import TimeOut
from .util import parse_address
import contextlib
import functools
import logging
import SimpleXMLRPCServer
import time

_log = logging.getLogger(__name__)


def authenticated(f):
    """Decorator to express that authentication is required."""
    @functools.wraps(f)
    def wrapper(self, cookie, *args):
        if cookie != self.cookie:
            raise MigrationError('authentication cookie mismatch')
        return f(self, *args)
    return wrapper


class IncomingServer(object):
    """Run an XML-RPC server to orchestrate a single migration."""

    finished = False

    def __init__(self, agent, timeout=330):
        self.agent = agent
        self.name = agent.name
        self.qemu = agent.qemu
        self.ceph = agent.ceph
        self.bind_address = parse_address(self.agent.migration_ctl_address)
        self.timeout = TimeOut(timeout, raise_on_timeout=False)
        self.consul = agent.consul

    _now = time.time

    @contextlib.contextmanager
    def inmigrate_service_registered(self):
        """Context manager for in-migration.

        Registers an inmigration service which keeps active as long as the
        with-block is executing.
        """
        svcname = 'vm-inmigrate-' + self.name
        self.consul.agent.service.register(
            svcname, address=self.bind_address[0], port=self.bind_address[1],
            ttl=self.timeout.remaining)
        try:
            yield
        finally:
            self.consul.agent.service.deregister(svcname)

    def run(self):
        s = SimpleXMLRPCServer.SimpleXMLRPCServer(
            self.bind_address, logRequests=False, allow_none=True)
        url = 'http://{}:{}/'.format(*self.bind_address)
        _log.info('%s: listening on %s', self.name, url)
        s.timeout = 1
        s.register_instance(IncomingAPI(self))
        s.register_introspection_functions()
        with self.inmigrate_service_registered():
            while self.timeout.tick():
                _log.debug('[server] %s: waiting (%ds remaining)', self.name,
                           int(self.timeout.remaining))
                s.handle_request()
                if self.finished:
                    break
        _log.info('%s: incoming migration returns %s', self.name,
                  self.finished)
        if self.finished == 'success':
            return 0
        else:
            self.qemu.destroy()
            return 1

    def extend_cutoff_time(self, timeout=60):
        self.timeout.cutoff = self._now() + timeout

    def prepare_incoming(self, args, config):
        self.qemu.args = args
        self.qemu.config = config
        try:
            return self.qemu.inmigrate()
        except Exception:
            _log.error('%s: incoming migration failed, releasing locks',
                       self.name)
            self.ceph.stop()
            raise

    def rescue(self):
        if not self.qemu.is_running():
            _log.warning('%s: trying to rescue, but VM is not online',
                         self.name)
            self.qemu.clean_run_files()
            self.ceph.stop()
            raise RuntimeError('rescue not possible - destroyed VM', self.name)
        try:
            _log.info('%s: rescue - assuming locks', self.name)
            self.ceph.lock()
        except Exception:
            _log.warning('%s: failed to acquire all locks', self.name)
            self.destroy()
            raise
        _log.info('%s: rescue succeeded, VM is running', self.name)

    def acquire_locks(self):
        self.ceph.lock()

    def finish_incoming(self):
        assert self.qemu.is_running()
        self.finished = 'success'

    def cancel(self):
        self.ceph.unlock()
        self.finished = 'canceled'

    def destroy(self):
        """Gets reliably rid of the VM."""
        _log.info('%s: self-destructing', self.name)
        self.finished = 'destroyed'
        try:
            self.qemu.destroy()
        except QemuNotRunning:
            pass
        self.qemu.clean_run_files()
        try:
            self.ceph.unlock()
        except Exception:
            pass


class IncomingAPI(object):

    def __init__(self, server):
        self.server = server
        self.cookie = server.agent.ceph.auth_cookie()

    @authenticated
    def ping(self):
        """Check connectivity and extend timeout."""
        _log.debug('[server] ping()')
        self.server.extend_cutoff_time()

    @authenticated
    def acquire_locks(self):
        _log.debug('[server] acquire_locks()')
        return self.server.acquire_locks()

    @authenticated
    def prepare_incoming(self, args, config):
        """Spawn KVM process ready to receive the VM.

        `args` and `config` should be the output of
        qemu.get_running_config() on the sending side.
        """
        _log.debug('[server] prepare_incoming()')
        return self.server.prepare_incoming(args, config)

    @authenticated
    def finish_incoming(self):
        _log.debug('[server] finish_incoming()')
        self.server.finish_incoming()

    @authenticated
    def rescue(self):
        """Incoming rescue."""
        _log.debug('[server] rescue()')
        return self.server.rescue()

    @authenticated
    def destroy(self):
        """Incoming destroy."""
        _log.debug('[server] destroy()')
        return self.server.destroy()

    @authenticated
    def cancel(self):
        _log.debug('[server] cancel()')
        self.server.cancel()



from ..agent import Agent, swap_size, tmp_size
import os
import pkg_resources
import pytest
import shutil
import subprocess
import traceback
import sys


@pytest.yield_fixture
def clean_environment():
    def clean():
        subprocess.call('pkill -f qemu', shell=True)
        subprocess.call('rbd rm test/test00.swap', shell=True)
        subprocess.call('rbd rm test/test00.root', shell=True)
        subprocess.call('rbd rm test/test00.tmp', shell=True)
    clean()
    yield
    clean()


@pytest.yield_fixture
def vm(clean_environment):
    fixtures = pkg_resources.resource_filename(__name__, 'fixtures')
    shutil.copy(fixtures + '/simplevm.yaml', '/etc/qemu/vm/simplevm.cfg')
    vm = Agent('simplevm')
    vm.timeout_graceful = 1
    vm.__enter__()
    yield vm
    exc_info = sys.exc_info()
    vm.__exit__(*exc_info)
    if len(exc_info):
        print(traceback.print_tb(exc_info[2]))
    os.unlink('/etc/qemu/vm/simplevm.cfg')


def test_simple_vm_lifecycle_start_stop(vm, capfd):

    def status():
        capfd.readouterr()
        vm.status()
        out, err = capfd.readouterr()
        return out

    assert status() == 'offline\n'

    vm.start()

    out, err = capfd.readouterr()
    assert """\
/usr/local/sbin/create-vm -I test00
rbd --id "admin" map "test/test00.tmp"
mkfs -F -q -m 1 -t ext4 -L "tmp" "/dev/rbd/test/test00.tmp"
tune2fs -e remount-ro "/dev/rbd/test/test00.tmp"
tune2fs 1.42.9 (4-Feb-2014)
Setting error behavior to 2
rbd --id "admin" unmap "/dev/rbd/test/test00.tmp"
rbd --id "admin" map "test/test00.swap"
mkswap -f -L "swap" "/dev/rbd/test/test00.swap"
""" in out

    assert """\
rbd --id "admin" unmap "/dev/rbd/test/test00.swap"
""" in out

    assert status() == """\
online
lock: test00.root@host1
lock: test00.swap@host1
lock: test00.tmp@host1
"""

    vm.stop()
    assert status() == 'offline\n'


def test_simple_vm_lifecycle_ensure_going_offline(vm, capsys, caplog):
    def status():
        capsys.readouterr()
        vm.status()
        out, err = capsys.readouterr()
        return out

    assert status() == 'offline\n'

    vm.ensure()

    out, err = capsys.readouterr()
    assert 'VM test00 should be running here' in caplog.text()
    assert status() == """\
online
lock: test00.root@host1
lock: test00.swap@host1
lock: test00.tmp@host1
"""

    vm.cfg['online'] = False
    vm.save_enc()
    vm.ensure()
    assert status() == 'offline\n'


def test_vm_not_running_here(vm, capsys):
    def status():
        capsys.readouterr()
        vm.status()
        out, err = capsys.readouterr()
        return out

    assert status() == 'offline\n'

    vm.cfg['kvm_host'] = 'otherhost'
    vm.ensure()
    assert status() == 'offline\n'


def test_crashed_vm_clean_restart(vm, capsys):
    def status():
        capsys.readouterr()
        vm.status()
        out, _err = capsys.readouterr()
        return out

    assert status() == 'offline\n'

    vm.ensure()

    assert status() == """\
online
lock: test00.root@host1
lock: test00.swap@host1
lock: test00.tmp@host1
"""

    vm.qemu.proc().kill()
    vm.qemu.proc().wait(2)
    assert status() == """\
offline
lock: test00.root@host1
lock: test00.swap@host1
lock: test00.tmp@host1
"""

    vm.ensure()
    assert status() == """\
online
lock: test00.root@host1
lock: test00.swap@host1
lock: test00.tmp@host1
"""

    vm.stop()
    assert status() == 'offline\n'


def test_do_not_clean_up_crashed_vm_that_doesnt_get_restarted(vm):
    vm.ensure()
    assert vm.qemu.is_running() is True
    vm.qemu.proc().kill()
    vm.qemu.proc().wait(2)
    assert vm.ceph.locked_by_me() is True
    vm.cfg['online'] = False
    vm.ensure()
    # We don't really know what's going on here, so, yeah, don't touch it.
    assert vm.ceph.locked_by_me() is True


def test_vm_swapsize():
    assert swap_size(256) == 1 * 1024 ** 3
    assert swap_size(512) == 1 * 1024 ** 3
    assert swap_size(768) == 1 * 1024 ** 3
    assert swap_size(1024) == 1 * 1024 ** 3
    assert swap_size(2048) == 1 * 1024 ** 3
    assert swap_size(4096) == 2 * 1024 ** 3


def test_vm_tmpsize():
    assert tmp_size(5) == 5120 * 1024 ** 2
    assert tmp_size(10) == 5120 * 1024 ** 2
    assert tmp_size(50) == 5120 * 1024 ** 2
    assert tmp_size(100) == 10240 * 1024 ** 2
    assert tmp_size(200) == 20480 * 1024 ** 2


def test_vm_migration():
    subprocess.check_call('/vagrant/test-migration.sh', shell=True)
